from typing import List

from Interpreter.Repo import Repo
from Interpreter.Utils import Utils
from Models.Token.Token import Token
from Models.Token.TokenTypes import TokenTypes, UnresolvedTokenTypes


class Lexer:

    def __init__(self, repo: Repo, utils: Utils) -> None:
        self.__repo = repo
        self.__utils = utils

        self.__is_ml_comment = False

    def __add_tok(self, tokens: List[Token], value: str, input_token_type: UnresolvedTokenTypes) -> None:
        """
        Creates token object
        :param value:
        :param input_token_type:
        :return:
        """
        token_type: TokenTypes = TokenTypes.Empty

        if value != '':
            if input_token_type == UnresolvedTokenTypes.Word:
                if value in self.__repo.keywords:
                    token_type = TokenTypes.Keyword

                elif value in self.__repo.datatypes:
                    token_type = TokenTypes.Type

                elif value in self.__repo.bool_values:
                    token_type = TokenTypes.Bool

                elif value == UnresolvedTokenTypes.Empty:
                    token_type = TokenTypes.Empty

                else:
                    token_type = TokenTypes.Name

            elif input_token_type == UnresolvedTokenTypes.Number:
                token_type = TokenTypes.Number

            elif input_token_type == UnresolvedTokenTypes.String:
                token_type = TokenTypes.String

            elif input_token_type == UnresolvedTokenTypes.Expression_Character:
                token_type = TokenTypes.Expression_Character

            elif input_token_type == UnresolvedTokenTypes.Bool_Expression_Character:
                token_type = TokenTypes.Bool_Expression_Character

            elif input_token_type == UnresolvedTokenTypes.Format_Character:
                token_type = TokenTypes.Format_Character

            tokens.append(Token(token_type, value))

    def tokenize(self, line: str) -> List[Token]:
        """
        Creates token list from code line
        :param line:
        :return:
        """
        tokens: List[Token] = []
        word = ''
        ol_comment = False
        is_string1 = False  # 'hello'
        is_string2 = False  # "hello"
        is_number = False
        is_expr_char = False

        for i in range(0, len(line)):
            c = line[i]
            # ignore comments and spaces
            if not ol_comment and not self.__is_ml_comment:
                # comment filtering
                if c == '#' and not is_string1 and not is_string2:
                    ol_comment = True

                elif line[i - 1] == '/' and c == '/':
                    ol_comment = True

                elif line[i - 1] == '/' and c == '*':
                    self.__is_ml_comment = True
                    i += 2

                # end of number
                elif not c.isdigit() and c != '.' and is_number:
                    self.__add_tok(tokens, word, UnresolvedTokenTypes.Number)
                    local_tokens = self.tokenize(c)
                    for local_token in local_tokens:
                        tokens.append(local_token)

                    word = ''
                    is_number = False

                # end of expression char
                elif c not in self.__repo.expr_chars and is_expr_char:
                    self.__add_tok(tokens, word, UnresolvedTokenTypes.Expression_Character)
                    word = ''
                    is_expr_char = False

                # begin of is_string1
                elif c == '\'' and not is_string1:
                    is_string1 = True
                    word = ''

                # end of is_string1
                elif c == '\'' and is_string1:
                    is_string1 = False
                    self.__add_tok(tokens, word, UnresolvedTokenTypes.String)
                    word = ''

                # begin of is_string2
                elif c == '\"' and not is_string2:
                    is_string2 = True
                    word = ''

                # end of is_string2
                elif c == '\"' and is_string2:
                    is_string2 = False
                    self.__add_tok(tokens, word, UnresolvedTokenTypes.String)
                    word = ''

                # format char
                elif c in self.__repo.format_chars:
                    self.__add_tok(tokens, word, UnresolvedTokenTypes.Word)
                    self.__add_tok(tokens, c, UnresolvedTokenTypes.Format_Character)
                    word = ''

                # begin of number
                elif c.isdigit() and not is_number and word == '':
                    word += c
                    is_number = True

                # continue number
                elif (c.isdigit() or c == '.') and is_number:
                    word += c

                # begin expression char
                elif c in self.__repo.expr_chars and not is_expr_char:
                    word += c
                    is_expr_char = True

                # continue expression char
                elif c in self.__repo.expr_chars and is_expr_char:
                    word += c

                # bool expression char
                elif c in self.__repo.bool_expr_chars:
                    self.__add_tok(tokens, word, UnresolvedTokenTypes.Word)
                    self.__add_tok(tokens, c, UnresolvedTokenTypes.Bool_Expression_Character)
                    word = ''

                # end of word
                elif c == ' ' and not is_string1 and not is_string2 or c == '\n':
                    self.__add_tok(tokens, word, UnresolvedTokenTypes.Word)
                    word = ''

                else:
                    word += c

            if c == '\n' and ol_comment:
                ol_comment = False

            if line[i - 1] == '*' and c == '/':
                self.__is_ml_comment = False

        return tokens
